@@ -17,8 +17,8 @@ logger.setLevel(logging.WARNING)
 
 # Constants for the school easy API
 
-EASY_API_SPEED = 0.2  # (m/s)
-EASY_API_YAW_RATE = 0.5  # (rad/s)
+EASY_API_DEFAULT_SPEED = 0.2  # (m/s)
+EASY_API_DEFAULT_YAW_RATE = 0.5  # (rad/s)
 EASY_API_ADDITIONAL_DURATION = 10.0  # (s)
 EASY_API_TAKEOFF_ALTITUDE = 0.5  # should be roughly the same as default takeoff altitude (m)
 EASY_API_MAX_ALTITUDE = 0.9
@@ -64,6 +64,8 @@ class Cognifly:
         self.recv_port_video = recv_port_video
         self.wait_for_first_obs = wait_for_first_obs
         self.wait_for_first_obs_sleep_duration = wait_for_first_obs_sleep_duration
+        self.ea_speed = EASY_API_DEFAULT_SPEED
+        self.ea_yaw_rate = EASY_API_DEFAULT_YAW_RATE
 
         self.udp_int = UDPInterface()
         self.tcp_video_int = TCPVideoInterface()
@@ -229,7 +231,6 @@ class Cognifly:
         """
         timeout_t = time.time() + max_duration
         while True:
-            time.sleep(granularity)
             self._lock.acquire()
             c = self.__wait_done
             self._lock.release()
@@ -240,6 +241,21 @@ class Cognifly:
                     self.__wait_done = False
                     self._lock.release()
                 break
+            time.sleep(granularity)
+
+    def wait_acknowledgement(self, max_duration=1.0, granularity=0.05):
+        """
+        sleeps until an acknowledgement is received from the drone or max_duration is elapsed
+        """
+        timeout_t = time.time() + max_duration
+        while True:
+            self._lock.acquire()
+            c = self.__last_received_id < self.__last_sent_id
+            self._lock.release()
+            t = time.time()
+            if not c or t >= timeout_t:
+                break
+            time.sleep(granularity)
 
     # ==================================================================================================================
 
@@ -260,7 +276,7 @@ class Cognifly:
             time.sleep(sleep_duration)
             self._lock.acquire()
             c = self.__wait_ack_reset
-            if self.__wait_ack_reset:
+            if not self.__wait_ack_reset:
                 self.__wait_ack_reset = True  # for next reset
             self._lock.release()
         logger.info("Reset acknowlegement received.")
@@ -297,6 +313,8 @@ class Cognifly:
     def takeoff_nonblocking(self, altitude=None):
         """
         The drone takes off.
+
+        Do not use altitude (this currently refers to a FC throttle value and is likely to change in the future).
         """
         self.send(msg_type="ACT", msg=("TAKEOFF", altitude))
         self.time_takeoff = time.time()
@@ -348,6 +366,74 @@ class Cognifly:
         self._lock.release()
         self.send(msg_type="ACT", msg=(frame_str, x, y, z, yaw, max_velocity, max_yaw_rate, max_duration))
 
+    def set_pid_vel_x(self, k_p=None, k_i=None, k_d=None, timeout=1.0):
+        """
+        Sets the gains of the PID for the x velocity tracker.
+        Gains left as None are not affected.
+        Example of reasonable values:
+            k_p = 750.0
+            k_i = 400.0
+            k_d = 50.0
+        Args:
+            k_p: float (optional): proportional gain
+            k_i: float (optional): integral gain
+            k_d: float (optional): derivative gain
+            timeout: float (optional): timeout when waiting for confirmation from the drone
+        """
+        self.send(msg_type="PVX", msg=(k_p, k_i, k_d))
+        self.wait_acknowledgement(max_duration=timeout)
+
+    def set_pid_vel_y(self, k_p=None, k_i=None, k_d=None, timeout=1.0):
+        """
+        Sets the gains of the PID for the y velocity tracker.
+        Gains left as None are not affected.
+        Example of reasonable values:
+            k_p = 750.0
+            k_i = 400.0
+            k_d = 50.0
+        Args:
+            k_p: float (optional): proportional gain
+            k_i: float (optional): integral gain
+            k_d: float (optional): derivative gain
+            timeout: float (optional): timeout when waiting for confirmation from the drone
+        """
+        self.send(msg_type="PVY", msg=(k_p, k_i, k_d))
+        self.wait_acknowledgement(max_duration=timeout)
+
+    def set_pid_vel_z(self, k_p=None, k_i=None, k_d=None, timeout=1.0):
+        """
+        Sets the gains of the PID for the z velocity tracker.
+        Gains left as None are not affected.
+        Example of reasonable values:
+            k_p = 50.0
+            k_i = 20.0
+            k_d = 5.0
+        Args:
+            k_p: float (optional): proportional gain
+            k_i: float (optional): integral gain
+            k_d: float (optional): derivative gain
+            timeout: float (optional): timeout when waiting for confirmation from the drone
+        """
+        self.send(msg_type="PVZ", msg=(k_p, k_i, k_d))
+        self.wait_acknowledgement(max_duration=timeout)
+
+    def set_pid_vel_w(self, k_p=None, k_i=None, k_d=None, timeout=1.0):
+        """
+        Sets the gains of the PID for the yaw rate tracker.
+        Gains left as None are not affected.
+        Example of reasonable values:
+            k_p = 400.0
+            k_i = 200.0
+            k_d = 0.0
+        Args:
+            k_p: float (optional): proportional gain
+            k_i: float (optional): integral gain
+            k_d: float (optional): derivative gain
+            timeout: float (optional): timeout when waiting for confirmation from the drone
+        """
+        self.send(msg_type="PVW", msg=(k_p, k_i, k_d))
+        self.wait_acknowledgement(max_duration=timeout)
+
     # ==================================================================================================================
 
     # High-level API for school purpose
@@ -371,8 +457,8 @@ class Cognifly:
         self.easy_api_cur_z = altitude
         # this works (Z properly defined) but when done from the ground it is a bit violent and not very stable:
         self.set_position_nonblocking(x=0.0, y=0.0, z=self.easy_api_cur_z, yaw=0.0,
-                                      max_velocity=EASY_API_SPEED,
-                                      max_yaw_rate=EASY_API_YAW_RATE,
+                                      max_velocity=self.ea_speed,
+                                      max_yaw_rate=self.ea_yaw_rate,
                                       max_duration=alt_duration,
                                       relative=False)
         self.sleep_until_done(alt_duration)
@@ -391,10 +477,10 @@ class Cognifly:
         Goes forward by number of centimeters
         """
         val_m = val_cm / 100.0
-        duration = val_m / EASY_API_SPEED + EASY_API_ADDITIONAL_DURATION
+        duration = val_m / self.ea_speed + EASY_API_ADDITIONAL_DURATION
         self.set_position_nonblocking(x=val_m, y=0.0, z=self.easy_api_cur_z, yaw=None,
-                                      max_velocity=EASY_API_SPEED,
-                                      max_yaw_rate=EASY_API_YAW_RATE,
+                                      max_velocity=self.ea_speed,
+                                      max_yaw_rate=self.ea_yaw_rate,
                                       max_duration=duration,
                                       relative=True)
         self.sleep_until_done(duration)
@@ -404,10 +490,10 @@ class Cognifly:
         Goes backward by number of centimeters
         """
         val_m = val_cm / 100.0
-        duration = val_m / EASY_API_SPEED + EASY_API_ADDITIONAL_DURATION
+        duration = val_m / self.ea_speed + EASY_API_ADDITIONAL_DURATION
         self.set_position_nonblocking(x=-val_m, y=0.0, z=self.easy_api_cur_z, yaw=None,
-                                      max_velocity=EASY_API_SPEED,
-                                      max_yaw_rate=EASY_API_YAW_RATE,
+                                      max_velocity=self.ea_speed,
+                                      max_yaw_rate=self.ea_yaw_rate,
                                       max_duration=duration,
                                       relative=True)
         self.sleep_until_done(duration)
@@ -417,10 +503,10 @@ class Cognifly:
         Goes right by number of centimeters
         """
         val_m = val_cm / 100.0
-        duration = val_m / EASY_API_SPEED + EASY_API_ADDITIONAL_DURATION
+        duration = val_m / self.ea_speed + EASY_API_ADDITIONAL_DURATION
         self.set_position_nonblocking(x=0.0, y=val_m, z=self.easy_api_cur_z, yaw=None,
-                                      max_velocity=EASY_API_SPEED,
-                                      max_yaw_rate=EASY_API_YAW_RATE,
+                                      max_velocity=self.ea_speed,
+                                      max_yaw_rate=self.ea_yaw_rate,
                                       max_duration=duration,
                                       relative=True)
         self.sleep_until_done(duration)
@@ -430,10 +516,10 @@ class Cognifly:
         Goes left by number of centimeters
         """
         val_m = val_cm / 100.0
-        duration = val_m / EASY_API_SPEED + EASY_API_ADDITIONAL_DURATION
+        duration = val_m / self.ea_speed + EASY_API_ADDITIONAL_DURATION
         self.set_position_nonblocking(x=0.0, y=-val_m, z=self.easy_api_cur_z, yaw=None,
-                                      max_velocity=EASY_API_SPEED,
-                                      max_yaw_rate=EASY_API_YAW_RATE,
+                                      max_velocity=self.ea_speed,
+                                      max_yaw_rate=self.ea_yaw_rate,
                                       max_duration=duration,
                                       relative=True)
         self.sleep_until_done(duration)
@@ -443,12 +529,12 @@ class Cognifly:
         Goes up by number of centimeters
         """
         val_m = val_cm / 100.0
-        duration = val_m / EASY_API_SPEED + EASY_API_ADDITIONAL_DURATION
+        duration = val_m / self.ea_speed + EASY_API_ADDITIONAL_DURATION
         self.easy_api_cur_z += val_m
         self.easy_api_cur_z = clip(self.easy_api_cur_z, EASY_API_MIN_ALTITUDE, EASY_API_MAX_ALTITUDE)
         self.set_position_nonblocking(x=0.0, y=0.0, z=self.easy_api_cur_z, yaw=None,
-                                      max_velocity=EASY_API_SPEED,
-                                      max_yaw_rate=EASY_API_YAW_RATE,
+                                      max_velocity=self.ea_speed,
+                                      max_yaw_rate=self.ea_yaw_rate,
                                       max_duration=duration,
                                       relative=True)
         self.sleep_until_done(duration)
@@ -458,12 +544,12 @@ class Cognifly:
         Goes down by number of centimeters
         """
         val_m = val_cm / 100.0
-        duration = val_m / EASY_API_SPEED + EASY_API_ADDITIONAL_DURATION
+        duration = val_m / self.ea_speed + EASY_API_ADDITIONAL_DURATION
         self.easy_api_cur_z -= val_m
         self.easy_api_cur_z = clip(self.easy_api_cur_z, EASY_API_MIN_ALTITUDE, EASY_API_MAX_ALTITUDE)
         self.set_position_nonblocking(x=0.0, y=0.0, z=self.easy_api_cur_z, yaw=None,
-                                      max_velocity=EASY_API_SPEED,
-                                      max_yaw_rate=EASY_API_YAW_RATE,
+                                      max_velocity=self.ea_speed,
+                                      max_yaw_rate=self.ea_yaw_rate,
                                       max_duration=duration,
                                       relative=True)
         self.sleep_until_done(duration)
@@ -474,7 +560,7 @@ class Cognifly:
         Note: the drone takes the shortest path, direction is not guaranteed for angles close to 180.
         """
         # val_rad = max(val_deg * np.pi / 180.0, 0.0)
-        # duration = val_rad / EASY_API_YAW_RATE + EASY_API_ADDITIONAL_DURATION
+        # duration = val_rad / self.ea_yaw_rate + EASY_API_ADDITIONAL_DURATION
         # seq_angles = get_angle_sequence_rad(val_rad)
         # seq = []
         # for sa in seq_angles:
@@ -482,10 +568,10 @@ class Cognifly:
         # self.position_sequence(sequence=seq, max_duration=duration, relative=True)
 
         val_rad = np.pi if val_deg == 180 else (val_deg % 180) * np.pi / 180.0
-        duration = val_rad / EASY_API_YAW_RATE + EASY_API_ADDITIONAL_DURATION
+        duration = val_rad / self.ea_yaw_rate + EASY_API_ADDITIONAL_DURATION
         self.set_position_nonblocking(x=0.0, y=0.0, z=self.easy_api_cur_z, yaw=val_rad,
-                                      max_velocity=EASY_API_SPEED,
-                                      max_yaw_rate=EASY_API_YAW_RATE,
+                                      max_velocity=self.ea_speed,
+                                      max_yaw_rate=self.ea_yaw_rate,
                                       max_duration=duration,
                                       relative=True)
         self.sleep_until_done(duration)
@@ -496,7 +582,7 @@ class Cognifly:
         Note: the drone takes the shortest path, direction is not guaranteed for angles close to 180.
         """
         # val_rad = max(val_deg * np.pi / 180.0, 0.0)
-        # duration = val_rad / EASY_API_YAW_RATE + EASY_API_ADDITIONAL_DURATION
+        # duration = val_rad / self.ea_yaw_rate + EASY_API_ADDITIONAL_DURATION
         # val_rad = - val_rad
         # seq_angles = get_angle_sequence_rad(val_rad)
         # seq = []
@@ -505,10 +591,10 @@ class Cognifly:
         # self.position_sequence(sequence=seq, max_duration=duration, relative=True)
 
         val_rad = np.pi if val_deg == 180 else (val_deg % 180) * np.pi / 180.0
-        duration = val_rad / EASY_API_YAW_RATE + EASY_API_ADDITIONAL_DURATION
+        duration = val_rad / self.ea_yaw_rate + EASY_API_ADDITIONAL_DURATION
         self.set_position_nonblocking(x=0.0, y=0.0, z=self.easy_api_cur_z, yaw=-val_rad,
-                                      max_velocity=EASY_API_SPEED,
-                                      max_yaw_rate=EASY_API_YAW_RATE,
+                                      max_velocity=self.ea_speed,
+                                      max_yaw_rate=self.ea_yaw_rate,
                                       max_duration=duration,
                                       relative=True)
         self.sleep_until_done(duration)
@@ -524,8 +610,8 @@ class Cognifly:
         y = y * np.pi / 180.0
         self.easy_api_cur_z = clip(z, EASY_API_MIN_ALTITUDE, EASY_API_MAX_ALTITUDE)
         self.set_position_nonblocking(x=x, y=y, z=self.easy_api_cur_z, yaw=yaw,
-                                      max_velocity=EASY_API_SPEED,
-                                      max_yaw_rate=EASY_API_YAW_RATE,
+                                      max_velocity=self.ea_speed,
+                                      max_yaw_rate=self.ea_yaw_rate,
                                       max_duration=max_duration,
                                       relative=False)
         self.sleep_until_done(max_duration)
@@ -561,8 +647,8 @@ class Cognifly:
 
                 self.easy_api_cur_z = clip(z, EASY_API_MIN_ALTITUDE, EASY_API_MAX_ALTITUDE)
                 self.set_position_nonblocking(x=x, y=y, z=self.easy_api_cur_z, yaw=yaw,
-                                              max_velocity=EASY_API_SPEED if speed is None else speed,
-                                              max_yaw_rate=EASY_API_YAW_RATE if yaw_rate is None else yaw_rate,
+                                              max_velocity=self.ea_speed if speed is None else speed,
+                                              max_yaw_rate=self.ea_yaw_rate if yaw_rate is None else yaw_rate,
                                               max_duration=remaining_duration,
                                               relative=relative)
                 self.sleep_until_done(remaining_duration)
@@ -576,6 +662,22 @@ class Cognifly:
         """
         self.position_sequence(sequence=[[x1, y1, z1],[x2, y2, z2]], speed=speed)
 
+    def set_speed(self, speed):
+        """
+        Sets the maximum speed of the drone in the school API
+        Args:
+            speed: float: speed in cm/s; a reasonable value is 20.
+        """
+        self.ea_speed = speed / 100.0
+
+    def set_yaw_rate(self, yaw_rate):
+        """
+        Sets the maximum yaw rate of the drone in the school API
+        Args:
+            yaw_rate: float: yaw rate in deg/s; a reasonable value is 30.
+        """
+        self.ea_yaw_rate = yaw_rate * np.pi / 180.0
+
     # ==================================================================================================================
 
     # Methods common to both APIs:
@@ -584,15 +686,13 @@ class Cognifly:
         """
         Starts camera streaming.
         CAUTION: This will slow the frequency of the onboard controller down and may make the drone unstable!
-        In particular, we recommend not having the stream on during takeoff when using the school API.
         The highest resolution and fps are, the worst the influence on the onboard controller.
         The image transfers happens over TCP, which comes with a noticeable delay.
+        CAUTION: High fps may saturate the network in the presence of several drones.
         Args:
             resolution: Tuple(float: height, float: width): resolution of the captured images.
             fps: integer: maximum framerate (may not be attained),
-            display: boolean: whether to display the stream in an OpenCV window,
-                mostly for debugging purpose: this may output warnings or not work depending on opencv installation,
-                in particular, the window is created and updated in the receiver thread.
+            display: boolean: whether to display the stream in the GUI (requires GUI to be enabled),
             wait_first_frame: boolean: whether to sleep until the first frame is available.
         """
         self.tcp_video_int.start_receiver(self.recv_port_video, display)
@@ -718,7 +818,7 @@ class Cognifly:
     def get_yaw_rate(self):
         """
         Returns:
-            w: float: yaw rate in rad/s
+            w: float: yaw rate in deg/s
         """
         telemetry = self.get_telemetry()
         return telemetry[8] * 180.0 / np.pi
@@ -743,7 +843,7 @@ class Cognifly:
             speed: float: norm of the velocity, in cm/s
         """
         v_x, v_y, v_z = self.get_velocity()
-        return np.linalg.norm([v_x, v_y, v_z]) * 100
+        return np.linalg.norm([v_x, v_y, v_z])
 
 
 if __name__ == '__main__':
